/* In Ermangelung von LabVIEW wurde kurzerhand die Kommunikation in C++ geschrieben.
   Ist sowieso viel funktionssicherer!
 */
#include "main.h"

/********************************
 * Persistente Konfiguration:	*
 * Fensterposition, COM-Port	*
 ********************************/

Config config;

TCHAR listcaptions[4096];	// multi_sz

bool Config::load(const TCHAR*devicecode) {
 HKEY k1,k2;
 bool ret=false;
 if (!RegOpenKeyEx(HKEY_CURRENT_USER,TEXT("Software\\h#s"),0,KEY_READ,&k1)) {
  if (!RegOpenKeyEx(k1,TEXT("Linak"),0,KEY_READ,&k2)) {
   if (devicecode) {
    DWORD sz=sizeof listcaptions;
    if (!RegQueryValueEx(k2,devicecode,0,0,(BYTE*)listcaptions,&sz)
    && sz>80) ret=true;
   }else{
    DWORD sz=sizeof*this;
    if (!RegQueryValueEx(k2,TEXT("config"),0,0,(BYTE*)this,&sz)
    && sz==sizeof*this) ret=true;
   }
   RegCloseKey(k2);
  }
  RegCloseKey(k1);
 }
 return ret;
}

void regSetDescription(HKEY k,UINT id) {
 TCHAR s[64];
 int l=LoadString(0,id,s,elemof(s));
 RegSetValueEx(k,0,0,REG_SZ,(BYTE*)s,(l+1)*sizeof(TCHAR));
}

bool Config::save(const TCHAR*devicecode) const{
 HKEY k1,k2;
 DWORD d;
 bool ret=false;
 if (!RegCreateKeyEx(HKEY_CURRENT_USER,TEXT("Software\\h#s"),0,0,0,KEY_WRITE,0,&k1,&d)) {
  if (d==REG_CREATED_NEW_KEY) regSetDescription(k1,1);
  if (!RegCreateKeyEx(k1,TEXT("Linak"),0,0,0,KEY_WRITE,0,&k2,&d)) {
   if (d==REG_CREATED_NEW_KEY) regSetDescription(k2,2);
   if (devicecode) {
    const TCHAR*p=listcaptions; for (;*p;p+=lstrlen(p)+1);
    if (!RegSetValueEx(k2,devicecode,0,REG_MULTI_SZ,(BYTE*)listcaptions,DWORD(p-listcaptions+1)*sizeof(TCHAR))) ret=true;
   }else if (!RegSetValueEx(k2,TEXT("config"),0,REG_BINARY,(const BYTE*)this,sizeof*this)) ret=true;
   RegCloseKey(k2);
  }
  RegCloseKey(k1);
 }
 return ret;
}

/************************
 * Kommunikation und	*
 * Dialogelement-Update	*
 ************************/
#ifdef _DEBUG
void _cdecl debugOut(TCHAR*t,...) {
 TCHAR s[256];
 va_list va;
 va_start(va,t);
 _vsntprintf(s,elemof(s),t,va);
 va_end(va);
 OutputDebugString(s);
}
#endif

HWND hMainWnd;

static void FillComboWithComPorts(HWND hCombo) {
 ComboBox_ResetContent(hCombo);
 HDEVINFO devs=SetupDiGetClassDevs(const_cast<GUID*>(&GUID_DEVCLASS_PORTS),0,0,DIGCF_PRESENT);
 if (devs!=INVALID_HANDLE_VALUE) {
  SP_DEVINFO_DATA devinfo;
  devinfo.cbSize=sizeof devinfo;
  for (DWORD i=0; SetupDiEnumDeviceInfo(devs,i,&devinfo); i++) {
   TCHAR s[16];
   DWORD size=sizeof s;
   int k;
   HKEY hKey;
   if ((hKey=SetupDiOpenDevRegKey(devs,&devinfo,DICS_FLAG_GLOBAL,0,DIREG_DEV,KEY_READ))
     ==INVALID_HANDLE_VALUE) continue;
   if (!RegQueryValueEx(hKey,TEXT("PortName"),NULL,NULL,(LPBYTE)s,&size)
   && _stscanf(s,TEXT("COM%d"),&k)==1) {	// LPTs herauswerfen
    int idx=ComboBox_AddString(hCombo,s);	// PortName-Wert, so wie er ist
    ComboBox_SetItemData(hCombo,idx,--k);	// numerischen Wert extra setzen
    if (k==config.comnr) ComboBox_SetCurSel(hCombo,idx);
   }
   RegCloseKey(hKey);
  }
  SetupDiDestroyDeviceInfoList(devs);
 }
}

static void lvSetColumns(HWND lv,const TCHAR*th, const BYTE*w, int ncol) {
 RECT r;
 GetClientRect(lv,&r);
 LVCOLUMN lvc;
 lvc.mask=LVCF_TEXT|LVCF_WIDTH;
 int defw=(r.right-r.left)/ncol;	// Gleiche Breite als Vorgabe
 for (int i=0;i<ncol;i++,w++) {
  lvc.cx=*w?*w:defw;
  lvc.pszText=const_cast<TCHAR*>(th);
  ListView_InsertColumn(lv,i,&lvc);
  th+=lstrlen(th)+1;		// zum nchsten Teilstring
 }
 ListView_SetExtendedListViewStyle(lv,LVS_EX_GRIDLINES|LVS_EX_FULLROWSELECT);
}

static void lvSaveColWidths(HWND lv,BYTE*w) {
 int cols=Header_GetItemCount(ListView_GetHeader(lv));
 for (int i=0; i<cols; i++)
   *w++=BYTE(ListView_GetColumnWidth(lv,i));
}

static WNDPROC defEdit;

static LRESULT CALLBACK myEditProc(HWND Wnd, UINT msg, WPARAM wParam, LPARAM lParam) {
 switch (msg) {
  case WM_GETDLGCODE: return DLGC_WANTALLKEYS|CallWindowProc(defEdit,Wnd,msg,wParam,lParam);	// sonst kommt kein VK_RETURN
  case WM_KILLFOCUS: DestroyWindow(Wnd); break;
  case WM_CHAR: switch (wParam) {
   case VK_RETURN:
   case VK_ESCAPE: return 0;
  }break;
  case WM_KEYDOWN: switch (wParam) {
   case VK_RETURN: {
    HWND lv=GetParent(Wnd);
    int i=ListView_GetNextItem(lv,-1,LVNI_SELECTED);
    TCHAR s[64];
    Edit_GetText(Wnd,s,elemof(s));
    ListView_SetItemText(lv,i,1,s);
   }/*nobreak*/
   case VK_ESCAPE: SetFocus(GetParent(Wnd)); return 0;
  }break;
 }
 return CallWindowProc(defEdit,Wnd,msg,wParam,lParam);
}

/*******************************************
 * Grenvernderlicher Dialog mit Divider *
 *******************************************/
static void gwr(HWND W, HWND w, RECT&R) {
 GetWindowRect(w,&R);
 R.right-=R.left;	// jetzt Breite
 R.bottom-=R.top;	// jetzt Hhe
 if (W) ScreenToClient(W,(POINT*)&R);
}

static struct Sizer{
 int minL;
 POINT ptMin;		// Window-Minimum
 POINT szMin,szBefore;	// Client-Minimum und Vorher-Wert fr WM_SIZE
// int W() const {return minR-minL;}	// Sizer-Breite = gesamter Raum zwischen den beiden ListView
 int limitDiv(int div, int cx) const;	// <div> fr gegebene Breite <cx> einschrnken; cx muss >= szMin.cx sein!
 int mouseStartX;			// Mausposition bei WM_LBUTTONDOWN
 bool initialized() const {return !!ptMin.y;}
 void init();
 static int getDividerLeft();
 static int getDividerRight();
 static int getFixedTopRight();
 static void handleReposition(int dxl, int dxr, int dy);
 static bool hitDivider(int);
}sizer;

int Sizer::limitDiv(int div, int w) const{
 if (div<minL) div=minL;
 w-=szMin.x-minL;
 if (div>w) div=w;
 return div;
}

int Sizer::getDividerLeft() {
 RECT R;
 gwr(hMainWnd,GetPrevSibling(GetDlgItem(hMainWnd,11)),R);
 return R.left+R.right;	// derzeitiger rechter Rand der Parameter = Divider linker Rand
}

int Sizer::getDividerRight() {
 RECT R;
 gwr(hMainWnd,GetPrevSibling(GetDlgItem(hMainWnd,12)),R);
 return R.left;		// derzeitiger linker Rand der Werte = Divider rechter Rand
}

int Sizer::getFixedTopRight() {
 RECT R;
 gwr(hMainWnd,GetPrevSibling(GetDlgItem(hMainWnd,12)),R);
 return R.top;		// derzeitiger oberer Rand der Werte = Beginn variable Hhe rechts
}

bool Sizer::hitDivider(int x) {
 return getDividerLeft()<=x && x<sizer.getDividerRight();
}

void Sizer::init() {
 RECT R;
 gwr(0,hMainWnd,R);
 ptMin=*(POINT*)&R.right;
 GetClientRect(hMainWnd,&R);	// Initiale Fenstergre aus Ressource ist Minimum; in Pixel speichern
 szMin.x=R.right-R.left;
 szMin.y=R.bottom-R.top;
 szBefore=szMin;
 minL=getDividerLeft();
}

void Sizer::handleReposition(int dxl, int dxr, int dy) {
// dxl = Grennderung linker Teil
// dxr = Grennderung rechter Teil
// dy = Hhennderung (nur bei WM_SIZE), alles in Pixel
// Mae bereits geprft!?
 int div=getDividerLeft();	// derzeitiger rechter Rand der Parameter = Divider (linker Rand)
 int fix=getFixedTopRight();
 HDWP wpi=BeginDeferWindowPos(10);
 for (HWND w=GetFirstChild(hMainWnd);w;w=GetNextSibling(w)) {
  RECT R;
  gwr(hMainWnd,w,R);
  DWORD flags=SWP_NOMOVE|SWP_NOSIZE|SWP_NOZORDER;
  if (R.left<div) {	// linker Teil
   R.right+=dxl;
   R.bottom+=dy;
   flags&=~SWP_NOSIZE;	// skalieren
  }else{		// rechter Teil
   R.left+=dxl;
   flags&=~SWP_NOMOVE;	// Nur horizontal verschieben
   if (R.top+R.bottom>=fix) { // rechter unterer Teil
    R.right+=dxr;
    R.bottom+=dy;
    flags&=~SWP_NOSIZE;
   }	// skalieren
  }
  wpi=DeferWindowPos(wpi,w,0,R.left,R.top,R.right,R.bottom,flags);
 }
 EndDeferWindowPos(wpi);
}

static INT_PTR CALLBACK MainDlgProc(HWND Wnd, UINT msg, WPARAM wParam, LPARAM lParam) {
 switch (msg) {
  case WM_INITDIALOG: {
   hMainWnd=Wnd;
   HWND hParams=GetDlgItem(Wnd,11);
   HWND hValues=GetDlgItem(Wnd,12);
   sizer.init();
   if (config.load()) {
    WINDOWPLACEMENT wp;
    wp.length=sizeof wp;
    GetWindowPlacement(Wnd,&wp);
    config.rc.setRect(wp.rcNormalPosition);
    wp.showCmd=config.showcmd;
    SetWindowPlacement(Wnd,&wp);
// Problem: Was tun wenn minimized??
    if (config.showcmd!=SW_MINIMIZE) {
     RECT R;
     GetClientRect(Wnd,&R);
     int divStartX=sizer.getDividerLeft();
     int dxl=sizer.limitDiv(config.hdiv,R.right)-divStartX;
     if (dxl) sizer.handleReposition(dxl,-dxl,0);
    }
   }
   TCHAR s[64];
   LoadString(0,4,s,elemof(s));	// Name \0 Wert \0 Minimum \0 Maximum
   lvSetColumns(hParams,s,config.cwParams,4);
   lvSetColumns(hValues,s,config.cwValues,2);
   FillComboWithComPorts(GetDlgItem(Wnd,10));
   worker.start();
  }return TRUE;
  case WM_GETMINMAXINFO: if (sizer.initialized()) {
   MINMAXINFO&mmi=*(MINMAXINFO*)lParam;
   mmi.ptMinTrackSize=sizer.ptMin;
  }break;
  case WM_SIZE: if (wParam==SIZE_RESTORED ||wParam==SIZE_MAXIMIZED) {
   POINT sz={GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)};
   if (sizer.initialized()) {	// Von WM_INITDIALOG initialisiert?
    int olddiv=sizer.getDividerLeft();
    int newdiv=MulDiv(olddiv,sz.x,sizer.szBefore.x);	// proportional verschieben
    int dxl=sizer.limitDiv(newdiv,sz.x)-olddiv;
    sizer.handleReposition(dxl,sz.x-sizer.szBefore.x-dxl,sz.y-sizer.szBefore.y);
   }
   sizer.szBefore=sz;
  }break;
  case WM_MOUSEMOVE:{
   POINT pt={GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)};
   if (GetCapture()==Wnd) {
    RECT R;
    gwr(0,Wnd,R);
// Durch das Limitieren ergibt sich ein unendliches Capture,
// whrend der Teiler nicht illegal verschoben werden kann.
    int divStartX=sizer.getDividerLeft();
    int dxl=sizer.limitDiv(divStartX+(pt.x-sizer.mouseStartX),R.right)-divStartX;
    if (dxl) sizer.handleReposition(dxl,-dxl,0);
    sizer.mouseStartX=pt.x;
    goto setcurs;
   }else if (sizer.hitDivider(pt.x)) {
    setcurs:
    SetCursor(LoadCursor(0,IDC_SIZEWE));
   }
  }break;
  case WM_LBUTTONDOWN:{
   POINT pt={GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)};
   if (sizer.hitDivider(pt.x)) {
    sizer.mouseStartX=pt.x;
    SetCapture(Wnd);
   }
  }break;
  case WM_LBUTTONUP:{
   ReleaseCapture();
  }break;
  case WM_COMMAND: switch (wParam) {
//   case IDOK: readValues(); break;
   case IDCANCEL: {
    worker.stop();
    WINDOWPLACEMENT wp;
    wp.length=sizeof wp;
    GetWindowPlacement(Wnd,&wp);
    config.rc=wp.rcNormalPosition;
    config.showcmd=wp.showCmd;
    if (wp.showCmd!=SW_MINIMIZE) config.hdiv=sizer.getDividerLeft();
    lvSaveColWidths(GetDlgItem(Wnd,11),config.cwParams);
    lvSaveColWidths(GetDlgItem(Wnd,12),config.cwValues);
    config.save();
    EndDialog(Wnd,0); 
   }break;
//   case 3: readParams(); break;
   case MAKELONG(10,CBN_SELCHANGE): {
    int idx=ComboBox_GetCurSel((HWND)lParam);
    if (idx>=0) {	// sollte hier stets der Fall sein, da nicht editierbar
     config.comnr=BYTE(ComboBox_GetItemData((HWND)lParam,idx));
     worker.stop();
     worker.start();
    }
   }break;
  }break;
  case WM_NOTIFY: {
   NMHDR&hdr=*(NMHDR*)lParam;
   switch (hdr.code) {
    case LVN_ENDLABELEDIT: {
     NMLVDISPINFO&lvdi=*(NMLVDISPINFO*)lParam;
     SetWindowLongPtr(Wnd,DWLP_MSGRESULT,TRUE);
    }return TRUE;
    case LVN_ITEMACTIVATE: {
     NMITEMACTIVATE&ia = *(NMITEMACTIVATE*)lParam;
     if (hdr.idFrom==11 && ia.iSubItem==1) {
      TCHAR s[64];
      ListView_GetItemText(hdr.hwndFrom,ia.iItem,ia.iSubItem,s,elemof(s));
      RECT r;
      ListView_GetSubItemRect(hdr.hwndFrom,ia.iItem,ia.iSubItem,LVIR_BOUNDS,&r);
      r.right-=r.left;	// Breite
      r.bottom-=r.top;	// Hhe
      HWND hEdit=CreateWindow(TEXT("EDIT"),s,ES_AUTOHSCROLL|WS_CHILD|WS_BORDER|WS_VISIBLE,
        r.left,r.top,r.right,r.bottom,hdr.hwndFrom,(HMENU)1,0,0);
      defEdit=SubclassWindow(hEdit,myEditProc);
      HFONT f=GetWindowFont(hdr.hwndFrom);
      if (f) SetWindowFont(hEdit,f,true);
      Edit_SetSel(hEdit,0,-1);
      SetFocus(hEdit);
     }
    }break;
   }
  }break;
#if 0
  case WM_TIMER: {
   KillTimer(Wnd,wParam);
   switch (wParam) {
    case 1: TimerTask1(); break;	// ID-String
    case 2: TimerTask2(); break;	// Minima/Maxima
    case 3: TimerTask3(); break;	// Parameter
    case 4: TimerTask4(); break;	// Werte (fortwhrend einlesen)
   }
  }break;
#endif
 }
 return FALSE;
}

void WinMainCRTStartup() {
 ExitProcess(UINT(DialogBox(0,MAKEINTRESOURCE(1),0,MainDlgProc)));
}
